@@ -1,7 +1,10 @@
 //! Type inference for expressions.
 
-use std::iter::{repeat, repeat_with};
-use std::{mem, sync::Arc};
+use std::{
+    iter::{repeat, repeat_with},
+    mem,
+    sync::Arc,
+};
 
 use chalk_ir::{cast::Cast, fold::Shift, Mutability, TyVariableKind};
 use hir_def::{
@@ -15,7 +18,9 @@ use stdx::always;
 use syntax::ast::RangeOp;
 
 use crate::{
-    autoderef, consteval,
+    autoderef::{self, Autoderef},
+    consteval,
+    infer::coerce::CoerceMany,
     lower::lower_to_chalk_mutability,
     mapping::from_chalk,
     method_resolution, op,
@@ -56,15 +61,16 @@ impl<'a> InferenceContext<'a> {
     pub(super) fn infer_expr_coerce(&mut self, expr: ExprId, expected: &Expectation) -> Ty {
         let ty = self.infer_expr_inner(expr, expected);
         let ty = if let Some(target) = expected.only_has_type(&mut self.table) {
-            if !self.coerce(&ty, &target) {
-                self.result
-                    .type_mismatches
-                    .insert(expr.into(), TypeMismatch { expected: target, actual: ty.clone() });
-                // Return actual type when type mismatch.
-                // This is needed for diagnostic when return type mismatch.
-                ty
-            } else {
-                target
+            match self.coerce(Some(expr), &ty, &target) {
+                Ok(res) => res.value,
+                Err(_) => {
+                    self.result
+                        .type_mismatches
+                        .insert(expr.into(), TypeMismatch { expected: target, actual: ty.clone() });
+                    // Return actual type when type mismatch.
+                    // This is needed for diagnostic when return type mismatch.
+                    ty
+                }
             }
         } else {
             ty
@@ -124,31 +130,32 @@ impl<'a> InferenceContext<'a> {
         let body = Arc::clone(&self.body); // avoid borrow checker problem
         let ty = match &body[tgt_expr] {
             Expr::Missing => self.err_ty(),
-            Expr::If { condition, then_branch, else_branch } => {
+            &Expr::If { condition, then_branch, else_branch } => {
                 // if let is desugared to match, so this is always simple if
                 self.infer_expr(
-                    *condition,
+                    condition,
                     &Expectation::has_type(TyKind::Scalar(Scalar::Bool).intern(&Interner)),
                 );
 
                 let condition_diverges = mem::replace(&mut self.diverges, Diverges::Maybe);
                 let mut both_arms_diverge = Diverges::Always;
 
-                let mut result_ty = self.table.new_type_var();
-                let then_ty = self.infer_expr_inner(*then_branch, expected);
+                let result_ty = self.table.new_type_var();
+                let then_ty = self.infer_expr_inner(then_branch, expected);
                 both_arms_diverge &= mem::replace(&mut self.diverges, Diverges::Maybe);
-                result_ty = self.coerce_merge_branch(Some(*then_branch), &result_ty, &then_ty);
+                let mut coerce = CoerceMany::new(result_ty);
+                coerce.coerce(self, Some(then_branch), &then_ty);
                 let else_ty = match else_branch {
-                    Some(else_branch) => self.infer_expr_inner(*else_branch, expected),
+                    Some(else_branch) => self.infer_expr_inner(else_branch, expected),
                     None => TyBuilder::unit(),
                 };
                 both_arms_diverge &= self.diverges;
                 // FIXME: create a synthetic `else {}` so we have something to refer to here instead of None?
-                result_ty = self.coerce_merge_branch(*else_branch, &result_ty, &else_ty);
+                coerce.coerce(self, else_branch, &else_ty);
 
                 self.diverges = condition_diverges | both_arms_diverge;
 
-                result_ty
+                coerce.complete()
             }
             Expr::Block { statements, tail, label, id: _ } => {
                 let old_resolver = mem::replace(
@@ -160,19 +167,23 @@ impl<'a> InferenceContext<'a> {
                         let break_ty = self.table.new_type_var();
                         self.breakables.push(BreakableContext {
                             may_break: false,
-                            break_ty: break_ty.clone(),
+                            coerce: CoerceMany::new(break_ty.clone()),
                             label: label.map(|label| self.body[label].name.clone()),
                         });
-                        let ty =
-                            self.infer_block(statements, *tail, &Expectation::has_type(break_ty));
+                        let ty = self.infer_block(
+                            tgt_expr,
+                            statements,
+                            *tail,
+                            &Expectation::has_type(break_ty),
+                        );
                         let ctxt = self.breakables.pop().expect("breakable stack broken");
                         if ctxt.may_break {
-                            ctxt.break_ty
+                            ctxt.coerce.complete()
                         } else {
                             ty
                         }
                     }
-                    None => self.infer_block(statements, *tail, expected),
+                    None => self.infer_block(tgt_expr, statements, *tail, expected),
                 };
                 self.resolver = old_resolver;
                 ty
@@ -185,7 +196,7 @@ impl<'a> InferenceContext<'a> {
             }
             Expr::Async { body } => {
                 // Use the first type parameter as the output type of future.
-                // existenail type AsyncBlockImplTrait<InnerType>: Future<Output = InnerType>
+                // existential type AsyncBlockImplTrait<InnerType>: Future<Output = InnerType>
                 let inner_ty = self.infer_expr(*body, &Expectation::none());
                 let impl_trait_id = crate::ImplTraitId::AsyncBlockTypeImplTrait(self.owner, *body);
                 let opaque_ty_id = self.db.intern_impl_trait_id(impl_trait_id).into();
@@ -195,18 +206,16 @@ impl<'a> InferenceContext<'a> {
             Expr::Loop { body, label } => {
                 self.breakables.push(BreakableContext {
                     may_break: false,
-                    break_ty: self.table.new_type_var(),
+                    coerce: CoerceMany::new(self.table.new_type_var()),
                     label: label.map(|label| self.body[label].name.clone()),
                 });
                 self.infer_expr(*body, &Expectation::has_type(TyBuilder::unit()));
 
                 let ctxt = self.breakables.pop().expect("breakable stack broken");
-                if ctxt.may_break {
-                    self.diverges = Diverges::Maybe;
-                }
 
                 if ctxt.may_break {
-                    ctxt.break_ty
+                    self.diverges = Diverges::Maybe;
+                    ctxt.coerce.complete()
                 } else {
                     TyKind::Never.intern(&Interner)
                 }
@@ -214,7 +223,7 @@ impl<'a> InferenceContext<'a> {
             Expr::While { condition, body, label } => {
                 self.breakables.push(BreakableContext {
                     may_break: false,
-                    break_ty: self.err_ty(),
+                    coerce: CoerceMany::new(self.err_ty()),
                     label: label.map(|label| self.body[label].name.clone()),
                 });
                 // while let is desugared to a match loop, so this is always simple while
@@ -233,7 +242,7 @@ impl<'a> InferenceContext<'a> {
 
                 self.breakables.push(BreakableContext {
                     may_break: false,
-                    break_ty: self.err_ty(),
+                    coerce: CoerceMany::new(self.err_ty()),
                     label: label.map(|label| self.body[label].name.clone()),
                 });
                 let pat_ty =
@@ -278,15 +287,18 @@ impl<'a> InferenceContext<'a> {
                 .intern(&Interner);
                 let closure_id = self.db.intern_closure((self.owner, tgt_expr)).into();
                 let closure_ty =
-                    TyKind::Closure(closure_id, Substitution::from1(&Interner, sig_ty))
+                    TyKind::Closure(closure_id, Substitution::from1(&Interner, sig_ty.clone()))
                         .intern(&Interner);
 
                 // Eagerly try to relate the closure type with the expected
                 // type, otherwise we often won't have enough information to
                 // infer the body.
-                if let Some(t) = expected.only_has_type(&mut self.table) {
-                    self.coerce(&closure_ty, &t);
-                }
+                self.deduce_closure_type_from_expectations(
+                    tgt_expr,
+                    &closure_ty,
+                    &sig_ty,
+                    expected,
+                );
 
                 // Now go through the argument patterns
                 for (arg_pat, arg_ty) in args.iter().zip(sig_tys) {
@@ -306,7 +318,7 @@ impl<'a> InferenceContext<'a> {
             Expr::Call { callee, args } => {
                 let callee_ty = self.infer_expr(*callee, &Expectation::none());
                 let canonicalized = self.canonicalize(callee_ty.clone());
-                let mut derefs = autoderef(
+                let mut derefs = Autoderef::new(
                     self.db,
                     self.resolver.krate(),
                     InEnvironment {
@@ -314,14 +326,19 @@ impl<'a> InferenceContext<'a> {
                         environment: self.table.trait_env.env.clone(),
                     },
                 );
-                let (param_tys, ret_ty): (Vec<Ty>, Ty) = derefs
-                    .find_map(|callee_deref_ty| {
-                        self.callable_sig(
-                            &canonicalized.decanonicalize_ty(callee_deref_ty.value),
-                            args.len(),
-                        )
-                    })
-                    .unwrap_or((Vec::new(), self.err_ty()));
+                let res = derefs.by_ref().find_map(|(callee_deref_ty, _)| {
+                    self.callable_sig(
+                        &canonicalized.decanonicalize_ty(callee_deref_ty.value),
+                        args.len(),
+                    )
+                });
+                let (param_tys, ret_ty): (Vec<Ty>, Ty) = match res {
+                    Some(res) => {
+                        self.write_expr_adj(*callee, self.auto_deref_adjust_steps(&derefs));
+                        res
+                    }
+                    None => (Vec::new(), self.err_ty()),
+                };
                 self.register_obligations_for_call(&callee_ty);
                 self.check_call_arguments(args, &param_tys);
                 self.normalize_associated_types_in(ret_ty)
@@ -333,7 +350,7 @@ impl<'a> InferenceContext<'a> {
 
                 let expected = expected.adjust_for_branches(&mut self.table);
 
-                let mut result_ty = if arms.is_empty() {
+                let result_ty = if arms.is_empty() {
                     TyKind::Never.intern(&Interner)
                 } else {
                     match &expected {
@@ -341,6 +358,7 @@ impl<'a> InferenceContext<'a> {
                         _ => self.table.new_type_var(),
                     }
                 };
+                let mut coerce = CoerceMany::new(result_ty);
 
                 let matchee_diverges = self.diverges;
                 let mut all_arms_diverge = Diverges::Always;
@@ -357,12 +375,12 @@ impl<'a> InferenceContext<'a> {
 
                     let arm_ty = self.infer_expr_inner(arm.expr, &expected);
                     all_arms_diverge &= self.diverges;
-                    result_ty = self.coerce_merge_branch(Some(arm.expr), &result_ty, &arm_ty);
+                    coerce.coerce(self, Some(arm.expr), &arm_ty);
                 }
 
                 self.diverges = matchee_diverges | all_arms_diverge;
 
-                result_ty
+                coerce.complete()
             }
             Expr::Path(p) => {
                 // FIXME this could be more efficient...
@@ -371,30 +389,35 @@ impl<'a> InferenceContext<'a> {
             }
             Expr::Continue { .. } => TyKind::Never.intern(&Interner),
             Expr::Break { expr, label } => {
-                let last_ty =
-                    if let Some(ctxt) = find_breakable(&mut self.breakables, label.as_ref()) {
-                        ctxt.break_ty.clone()
-                    } else {
-                        self.err_ty()
-                    };
+                let mut coerce = match find_breakable(&mut self.breakables, label.as_ref()) {
+                    Some(ctxt) => {
+                        // avoiding the borrowck
+                        mem::replace(
+                            &mut ctxt.coerce,
+                            CoerceMany::new(self.result.standard_types.unknown.clone()),
+                        )
+                    }
+                    None => CoerceMany::new(self.result.standard_types.unknown.clone()),
+                };
 
-                let val_ty = if let Some(expr) = expr {
-                    self.infer_expr(*expr, &Expectation::none())
+                let val_ty = if let Some(expr) = *expr {
+                    self.infer_expr(expr, &Expectation::none())
                 } else {
                     TyBuilder::unit()
                 };
 
                 // FIXME: create a synthetic `()` during lowering so we have something to refer to here?
-                let merged_type = self.coerce_merge_branch(*expr, &last_ty, &val_ty);
+                coerce.coerce(self, *expr, &val_ty);
 
                 if let Some(ctxt) = find_breakable(&mut self.breakables, label.as_ref()) {
-                    ctxt.break_ty = merged_type;
+                    ctxt.coerce = coerce;
                     ctxt.may_break = true;
                 } else {
                     self.push_diagnostic(InferenceDiagnostic::BreakOutsideOfLoop {
                         expr: tgt_expr,
                     });
-                }
+                };
+
                 TyKind::Never.intern(&Interner)
             }
             Expr::Return { expr } => {
@@ -402,7 +425,7 @@ impl<'a> InferenceContext<'a> {
                     self.infer_expr_coerce(*expr, &Expectation::has_type(self.return_ty.clone()));
                 } else {
                     let unit = TyBuilder::unit();
-                    self.coerce(&unit, &self.return_ty.clone());
+                    let _ = self.coerce(Some(tgt_expr), &unit, &self.return_ty.clone());
                 }
                 TyKind::Never.intern(&Interner)
             }
@@ -453,15 +476,16 @@ impl<'a> InferenceContext<'a> {
             Expr::Field { expr, name } => {
                 let receiver_ty = self.infer_expr_inner(*expr, &Expectation::none());
                 let canonicalized = self.canonicalize(receiver_ty);
-                let ty = autoderef::autoderef(
+
+                let mut autoderef = Autoderef::new(
                     self.db,
                     self.resolver.krate(),
                     InEnvironment {
                         goal: canonicalized.value.clone(),
                         environment: self.trait_env.env.clone(),
                     },
-                )
-                .find_map(|derefed_ty| {
+                );
+                let ty = autoderef.by_ref().find_map(|(derefed_ty, _)| {
                     let def_db = self.db.upcast();
                     let module = self.resolver.module();
                     let is_visible = |field_id: &FieldId| {
@@ -510,8 +534,14 @@ impl<'a> InferenceContext<'a> {
                         }
                         _ => None,
                     }
-                })
-                .unwrap_or_else(|| self.err_ty());
+                });
+                let ty = match ty {
+                    Some(ty) => {
+                        self.write_expr_adj(*expr, self.auto_deref_adjust_steps(&autoderef));
+                        ty
+                    }
+                    None => self.err_ty(),
+                };
                 let ty = self.insert_type_vars(ty);
                 self.normalize_associated_types_in(ty)
             }
@@ -736,33 +766,32 @@ impl<'a> InferenceContext<'a> {
                         Some(TyKind::Array(st, _) | TyKind::Slice(st)) => st.clone(),
                         _ => self.table.new_type_var(),
                     };
+                let mut coerce = CoerceMany::new(elem_ty.clone());
 
+                let expected = Expectation::has_type(elem_ty.clone());
                 let len = match array {
                     Array::ElementList(items) => {
-                        for expr in items.iter() {
-                            // FIXME: use CoerceMany (coerce_merge_branch)
-                            self.infer_expr_coerce(*expr, &Expectation::has_type(elem_ty.clone()));
+                        for &expr in items.iter() {
+                            let cur_elem_ty = self.infer_expr_inner(expr, &expected);
+                            coerce.coerce(self, Some(expr), &cur_elem_ty);
                         }
                         Some(items.len() as u64)
                     }
-                    Array::Repeat { initializer, repeat } => {
-                        self.infer_expr_coerce(
-                            *initializer,
-                            &Expectation::has_type(elem_ty.clone()),
-                        );
+                    &Array::Repeat { initializer, repeat } => {
+                        self.infer_expr_coerce(initializer, &Expectation::has_type(elem_ty));
                         self.infer_expr(
-                            *repeat,
+                            repeat,
                             &Expectation::has_type(
                                 TyKind::Scalar(Scalar::Uint(UintTy::Usize)).intern(&Interner),
                             ),
                         );
 
-                        let repeat_expr = &self.body.exprs[*repeat];
+                        let repeat_expr = &self.body.exprs[repeat];
                         consteval::eval_usize(repeat_expr)
                     }
                 };
 
-                TyKind::Array(elem_ty, consteval::usize_const(len)).intern(&Interner)
+                TyKind::Array(coerce.complete(), consteval::usize_const(len)).intern(&Interner)
             }
             Expr::Literal(lit) => match lit {
                 Literal::Bool(..) => TyKind::Scalar(Scalar::Bool).intern(&Interner),
@@ -811,6 +840,7 @@ impl<'a> InferenceContext<'a> {
 
     fn infer_block(
         &mut self,
+        expr: ExprId,
         statements: &[Statement],
         tail: Option<ExprId>,
         expected: &Expectation,
@@ -857,7 +887,7 @@ impl<'a> InferenceContext<'a> {
                 self.table.new_maybe_never_var()
             } else {
                 if let Some(t) = expected.only_has_type(&mut self.table) {
-                    self.coerce(&TyBuilder::unit(), &t);
+                    let _ = self.coerce(Some(expr), &TyBuilder::unit(), &t);
                 }
                 TyBuilder::unit()
             }
