@@ -54,3 +54,16 @@ fn main() {
 "#,
     );
 }
+
+#[test]
+fn render_dyn_for_ty() {
+    // FIXME
+    check_types_source_code(
+        r#"
+trait Foo<'a> {}
+
+fn foo(foo: &dyn for<'a> Foo<'a>) {}
+    // ^^^ &dyn Foo
+"#,
+    );
+}
